@@ -2,11 +2,31 @@
 #     RAAD Functionality    #
 #############################
 
-from time import daylight
 from .core import *
 from .rparray import array
 from .event import *
 
+# Split the dataset in channels
+def split_channels(data,struct=NONVETO_STRUCT):
+    """Split the data based on their channels
+
+    Args:
+        data (_type_): Buffer data
+        struct (_type_, optional): Structure to decode them as. Defaults to NONVETO_STRUCT.
+
+    Returns:
+        channels: List of lists for all the channels
+    """
+    # Split the data based on their channels
+    channels    = []
+    idxs        = []
+    for channel in np.unique(data['channel']):
+        idx         = np.where(data['channel'] == channel)[0]
+        idxs.append(idx.copy())
+        channels.append(dict(zip(struct.keys(),[arr[idx] for arr in data.values()])))
+    
+    return channels,idxs
+
 # Print the closest lightnings
 def get_nearby_lightning(tgf,lightnings:array,threshold:float=1):
     """Given an array, or a single event object filter a raadpy array that contains lightnings within a threshold.
@@ -164,6 +184,83 @@ def download_lightnings(event_time:Time,threshold:float = 6*60,VERBOSE=True):
 
     return download_lightnings_range(event_time-threshold,event_time+threshold,VERBOSE=VERBOSE)
 
+
+# Recursively keep removing the Most Significant Bit, until you're below the mean threshold
+def correct_bit(x:int, MEAN:float, BITS:int=48):
+    """Recursively Correct a number for bit flips, by removing the most significant bit until it's below a threshold
+
+    Args:
+        x (int): An integer to correct
+        MEAN (float): The value to reach
+        BITS (int, optional): The maximum number of bits that the number can have. Defaults to 48.
+
+    Returns:
+        x (int): The corrected number
+    """
+    if abs(x) <= MEAN: 
+        return x
+    if x > 0: return correct_bit(x - get_msb(x,BITS),MEAN,BITS)
+    else: return correct_bit(x + get_msb(abs(x),BITS),MEAN,BITS)
+
+
+# Find pairs
+def find_pairs(diffs,MEAN,STD):
+    """Helper function to find the pairs of bit flips occuring in a timestring, Use invert_flips instead!
+    """
+    candidates  = np.array([])
+    idxs        = np.array([],dtype=int)
+    round_msb = lambda data,BITS: get_msb((3*abs(data)).astype(int) >> 1,BITS)*np.sign(data)
+
+    pairs = []    
+    for i,d in enumerate(round_msb(diffs,50)):
+        # If this is a candidate for a thing 
+        if d < -MEAN - 3*STD:
+            candidates = np.append(candidates,[d])
+            idxs       = np.append(idxs      ,[i])
+        
+        if len(candidates) > 0:
+            # candidates += d
+            idx = np.where(abs(candidates + d) == 0)[0]
+            if len(idx) > 0:
+                index = idx[0]
+                pairs.append((idxs[index],i))
+                candidates = np.delete(candidates,[index,int(-1)],axis=0)
+                idxs       = np.delete(idxs      ,[index,int(-1)],axis=0)
+
+    return np.array(pairs)
+
+
+def invert_flips(timestamp:np.array,BITS:int=NONVETO_STRUCT['stimestamp']):
+    """Given a list of timestamps, find and correct the bit flips occured.
+
+    Args:
+        timestamp (np.array): The array of timestamps
+        BITS (dict, optional): The number of bits in the timestamp variable. Defaults to NONVETO_STRUCT['stimestamp'] = 48.
+
+    Returns:
+        timestamp (np.array): The corrected timestamp
+    """
+    # Get the gradient of the timestamp
+    timestamp_deltas = timestamp[1:] - timestamp[:-1]
+
+    # Get Mean and standard deviation
+    MEAN = np.mean(abs(timestamp_deltas))
+    STD  = np.std(abs(timestamp_deltas))
+
+    # Identify the pairs of points where you get bit flips
+    pairs = find_pairs(timestamp_deltas,MEAN,STD)
+    
+    # For each bit flip region
+    for pair in pairs:
+        ADD = 0
+        # For each point within the region
+        for i in range(*pair):
+            # Calculate a correction and apply it
+            ADD += correct_bit(int(timestamp_deltas[i]),MEAN,BITS=BITS) - timestamp_deltas[i]
+            timestamp[i+1] += ADD
+
+    return timestamp
+
 # We create a function that given a bytestring extracts the ith bit:
 def get_bit(i:int,string):
     '''
@@ -216,6 +313,8 @@ def get_dict(filename:str,struct=ORBIT_STRUCT,condition:str=None,MAX=None,STUPID
         condition (str, optional): If you want you can add a condition such as data['id_bit']==1 to filter the data as they're being loaded. Defaults to None.
         MAX (_type_, optional): Maximum number of lines to read, if None then read all of them. Defaults to None.
         STUPID (bool, optional): Should be set to True if you are reading VETO and NONVETO. Defaults to False.
+        VERIFY (bool, optional): Set to True to process error correction automatically, such as filtering the 2-byte error, and correcting for bit flips. Defaults to False
+        threshold (float,optional): The difference between two points in the timestamp that we can consider faulty as a fraction of the maximum number that the integer field can store. If threshold > 1 then it is considered as an absolute threhsold. Only used if VERIFY=True. Defaults to 5e-5
 
     Returns:
         data (dict): Dictionary with the decoded arrays of measurements
@@ -247,8 +346,6 @@ def get_dict(filename:str,struct=ORBIT_STRUCT,condition:str=None,MAX=None,STUPID
         else:
             THRESHOLD = threshold
 
-        # print("%.4e"%THRESHOLD)
-
     # Current byte index in the file
     curr = 0
     with tqdm(total=MAX,desc='Line: ') as pbar:
@@ -307,6 +404,20 @@ def get_dict(filename:str,struct=ORBIT_STRUCT,condition:str=None,MAX=None,STUPID
     if 'temperature' in struct.keys():
         data['temperature'] -= 55
         
+
+    # If we can do a bit flip verification perform it
+    if VERIFY:
+        # Split to channels
+        channels, cnt = split_channels(data,struct)
+        
+        # Apply correction to each channel
+        for channel in channels: channel['stimestamp'] = invert_flips(channel['stimestamp'],struct['stimestamp'])
+
+        # Put it back together
+        for i, channel in enumerate(channels):
+            for j, time in enumerate(channel['stimestamp']):
+                data['stimestamp'][cnt[i][j]] = time
+    
     # Return the dictionary
     return data
 
@@ -790,7 +901,7 @@ def log_expand(filename:str=None,text:str=None):
     # Collect the outputs of the commands
     decoded_log = [{
         'command':loglines[commands_idx[i]],
-        'output':loglines[commands_idx[i]+1:commands_idx[i+1]]
+        'output' :loglines[commands_idx[i]+1:commands_idx[i+1]]
         } for i in range(len(commands_idx)-1)]
 
     # Return
@@ -847,13 +958,13 @@ def log_metadata(decoded_log:list):
     }
 
     # Get the command list
-    commands = [row[0] for row in decoded_log]
+    commands = [row['command'] for row in decoded_log]
 
     # Find the start and end of the data acquisition
     # Index of start and end timestamps:
     start = [i for i in range(len(commands)) if "rtc read" in commands[i]]
-    if len(start) != 0: metadata['start_time']  = decoded_log[start[0] ][1][0][-21:-2]
-    if len(start) >= 2: metadata['end_time']    = decoded_log[start[-1]][1][0][-21:-2]
+    if len(start) != 0: metadata['start_time']  = decoded_log[start[0] ]['output'][0][-21:-2]
+    if len(start) >= 2: metadata['end_time']    = decoded_log[start[-1]]['output'][0][-21:-2]
 
     # Find the custom scenario commands for SiPM and PMT
     for num,payload in zip([12,13],['SiPM','PMT']):
@@ -992,111 +1103,6 @@ def desc_finder(line:str,cmdlist,outputs,i,time,failed_idx):
     
     return desc,time,status
 
-
-
-# Decode a logfile
-def decode_log(filename:str="../../../Data/Logs/light1-SD-1016-se-log.txt"):
-    """Take a log file and parse its commands to uncover thier timestamps
-
-    Args:
-        filename (str, optional): The filepath and filename of the log file. Defaults to "../../../Data/Logs/light1-SD-1016-se-log.txt".
-
-    Returns:
-        _type_: Log, commands, outputs, description, failed_idx, loglines_array
-    """
-
-    # Load the logfile
-    logfile = open(filename)
-    cmdlist = pd.read_csv("command_list.csv")
-
-    # Load the lines
-    loglines = logfile.readlines()
-
-    # Close the file
-    logfile.close()
-
-    # Create an array with the lines
-    commands    = []
-    outputs     = []
-    idx         = []
-    description = []
-    times       = []
-    fails       = []
-    failed_idx  = []
-
-    # Get commands and their indices
-    for i, line in enumerate(loglines):
-        if 'SE0>' in line:
-            commands.append(line)
-            idx.append(i)
-
-    # Check if the last command was empty
-    if loglines[-1] != 'SE0>':
-        commands.append('SE0>')
-        idx.append(len(loglines))
-
-    # Get the command output
-    for i in range(len(idx)-1):
-        out = []
-        if 'SE0>#' not in commands[i]:
-            for j in range(idx[i]+1,idx[i+1]):
-                out.append(loglines[j])
-        
-        outputs.append(out)
-
-    # Create the dictionary
-    log = [[command,output] for command,output in zip(commands,outputs)]
-
-    # Check which commands executed correcly
-  
-
-    # Find the commands that failed
-    for i in range(len(commands)-1):
-        for output in log[i][1]:
-            if 'FAIL' in output:
-                failed_idx.append(i)
-
-
-    # Initialize time at 0s
-    time = 0
-    
-    # Give commands and use the finder function to pull the description, time ran, and if the command failed.
-    for k, id in enumerate(idx):
-        c = commands[k]
-        cmd = c.split('\n')[0]
-        desc,time,failed = desc_finder(cmd,cmdlist,outputs,k,time,failed_idx)
-
-        if len(desc) != 0:
-            description.append(desc)
-            times.append(time)
-            fails.append(failed)
-    tempname = 'templog.csv'
-    fileDir = os.path.join('Defined_Logs')
-
-    if not os.path.exists(fileDir):
-        os.makedirs(fileDir)
-    
-    filePath = os.path.join(fileDir, tempname)
-
-    #Name the headers of the CSV File
-    header = ['status','time','description','ID_in_Graph']
-
-    loglines_array = []
-
-    #Clear the file and write from scratch
-    with open(filePath, 'w', newline='') as file:
-        writer = csv.writer(file)
-        writer.writerow(header)
-
-        for i in range(len(description)):
-            g = (fails[i], times[i-1], description[i][0],description[i][1])
-            loglines_array.append(g)
-            writer.writerow(g)
-
-    # Return everything else
-    return log, commands[:-1], outputs, description, failed_idx, loglines_array
-
-
 def log_line_timestamp(logline:list,time:float=0):
     """Given a log file line as the lines decoded by log_expand, and a time
     return the time increment for this line
@@ -1127,7 +1133,6 @@ def log_line_timestamp(logline:list,time:float=0):
     return time
 
 # Now we can define an error correction pass
-
 def find_closest(x,array):
     """Helper function that returns the two closest values to x in a list
 
@@ -1197,3 +1202,138 @@ def reorder_log(logfile:list):
 
 
     return logfile
+
+
+def log_with_timestamp(logfile:list):                        
+    """Given a log file from the function log_expand
+    return a list of log lines with the time increment of each line
+
+    Args:
+        logfile (list): a list of cubesat commands
+
+    Returns:
+        log_timestamp_list (list): a list of cubesat commands, each command with its timestamp
+    """
+    log_timestamp_list=[]                               # List to be returned later (cmnd, output, timestamp)
+    time = 0                                            # Set initial time to zero, this step will not be needed
+    for logline in logfile:
+        time = log_line_timestamp(logline,time)
+        log_timestamp_list.append ({
+            'command':      logline['command'],
+            'output':       logline['output'],
+            'timestamp':    time
+        })
+
+    return reorder_log(log_timestamp_list)
+
+
+def get_cmd_number(cmdline:str):
+    """Get the command number of a specific command line from the log file.
+
+    Args:
+        cmdline (str): The command lines as given by the log files. Example: "SE0>csp txrx 12 8 3000 0F".
+
+    Returns:
+        value (int): The number of the command on the buff 1 graphs. The value will be -1 for nonpayload cmds.
+    """
+    
+    value = -1                          # Initialize the value to be -1 so that if the function faces commands that are not in the CMND_LIST it will give it a -1 value
+    split_cmdline = cmdline.split(" ")  # Split the cmd by the spaces in between
+    if len(split_cmdline) > 1 and 'txrx' in split_cmdline[1]: # If the cmd is a payload cmd, is has txrx
+        if float(split_cmdline[2]) == 4:                      # If the cmd belongs to emergency poweroff then the message is not needed (message is split_cmdline[5])
+            value = CMND_LIST[split_cmdline[2]][split_cmdline[3]]['number_in_graph'] # Grab the cmd number from the CMND_LIST based on the node and port of the cmd ([split_cmdline[2]][split_cmdline[3]])
+        else:
+            value = CMND_LIST[split_cmdline[2]][split_cmdline[3]][split_cmdline[5]]['number_in_graph'] #If not an emergency poweroff use node, port, and message to find the cmd in the CMND_LIST
+    return value 
+        
+
+
+def send_sql_query_over_ssh(query:str):
+    """Receives SQL query and sends it to the raad@arneodolab.abudhabi.nyu.edu SQL server.
+
+    Args:
+        query (str): SQL Formatted Query
+
+    Returns:
+        data (pandas.dataframe): Dataframe of the data requested as formatted on the server.
+    """
+
+    # Set up ssh tunnel
+    # tunnel = paramiko.SSHClient()
+    # tunnel.set_missing_host_key_policy(paramiko.AutoAddPolicy())
+    # tunnel.connect('arneodolab.abudhabi.nyu.edu', 22,username='raad',password='nyuad123$',allow_agent=False,look_for_keys=False)
+    tunnel  = SSHTunnelForwarder(('arneodolab.abudhabi.nyu.edu', 22), ssh_password="nyuad123$", ssh_username="raad", remote_bind_address=('127.0.0.1', 3306), allow_agent=False,)
+    tunnel.start()
+
+    # Connect to PYSQl Host
+    conn    = pymysql.connect(host='127.0.0.1', user='raad', passwd="nyuad123$", port=tunnel.local_bind_port)
+    data    = pd.read_sql_query(query, conn)    # Request Data
+    tunnel.close()                              # Close tunnel
+    
+    return data
+
+
+def get_light1_position(starttime:Time, endtime:Time=None, n_events:int=None, SHORT_TIME:float=10):
+    """Give me a start and end time in astropy objects, I give you light-1 position. 
+    I can also interpolate if you give me anumber of events
+
+    Args:
+        starttime (Time): Start time, or single event time
+        endtime (Time, optional): End time. If left None, it assumes single position. Defaults to None.
+        n_events (int, optional): Number of points in case we interpolate. Defaults to None.
+        SHORT_TIME (float, optional): If you want a single event, give us some wiggle room to look around for it. Defaults to 10.
+
+    Raises:
+        Exception: Start Time and End Time are not astropy Time objects
+    
+    Returns:
+        locs (rp.array): Raadpy array with the locations of the satellite in this interval.
+    """
+    # Input processing on start time
+    if type(starttime) != Time:
+        try:
+            starttime = Time(starttime)
+        except:
+            raise Exception("Given timestamp is not an astropy Time object")       
+
+    # If endtime is not given, it means we want only one event, so we will interpolate around SHORT_TIME
+    if endtime is None:
+        starttime   -= TimeDelta(SHORT_TIME,format='sec')
+        endtime      = starttime + TimeDelta(2*SHORT_TIME,format='sec')
+        n_events    = -1
+    else: 
+        try:
+            endtime     = Time(endtime)
+        except:
+            raise Exception("Start time and End time are not astropy Time object")
+
+    # Request the data from sql for this time period
+    data =(send_sql_query_over_ssh("SELECT * FROM `LIGHT-1_Position_Data`.PositionData WHERE `Time (ModJDate)` BETWEEN " + str(starttime.to_value("mjd")) + " AND " + str(endtime.to_value("mjd")) + ";"))
+    
+    # Get the data from the dataframe
+    latitudes   = [i for i in data['Lat (deg)']]
+    longitudes  = [i for i in data['Lon (deg)']]
+    times       = [i for i in data['Time (ModJDate)']]
+    
+
+    # Interpolation if needed
+    if n_events is not None:
+        times_tmp  = np.linspace(starttime.to_value(format="mjd"), endtime.to_value(format="mjd"), n_events) if n_events > 0 else np.array([(starttime.mjd+endtime.mjd)/2])
+        latitudes  = np.interp(times_tmp, times, latitudes)
+        longitudes = np.interp(times_tmp, times, longitudes)
+        times = times_tmp
+    
+    # Create rp.array
+    locs = array(event_type="location")
+    for i in range(len(times)):
+        locs.append(event(
+            timestamp   = Time(times[i], format="mjd"),
+            longitude   = in_range(float(longitudes[i])),
+            latitude    = float(latitudes[i]),
+            detector_id = 'NA',
+            mission     = 'NanoAvionics',
+            time_format = "mjd",
+            event_type  = "cubesat-location"
+        ))
+
+    return(locs)
\ No newline at end of file
